/* comm/com.c */
/* channel comsystem - world-wide communications */

#include "externs.h"

int is_on_channel(dbref player, char *chan)
{
  char *p, *r, *s;

  /* search channel list for wordmatch */
  s=p=atr_get(player, A_CHANNEL);
  while((r=parse_up(&s, ' ')))
    if(!strcmp(r, chan))
      return (r-p)+1;
  return 0;
}

void com_send(dbref player, char *channel, char *message)
{
  DESC *d;
  char buf[8192];

  /* build message text */
  sprintf(buf, "%s[%s] %s%s", textcolor(7, default_color), channel, message,
          textcolor(default_color, 7));

  /* send message to all on channel */
  for(d=descriptor_list;d;d=d->next)
    if((d->flags & C_CONNECTED) && !IS(db[d->player].location, TYPE_ROOM,
       ROOM_XZONE) && is_on_channel(d->player, channel))
      notify(d->player, "%s", buf);
}

void com_who(dbref player, char *channel)
{
  DESC *d;
  int count=0;

  /* list who's on a particular channel */
  for(d=descriptor_list;d;d=d->next)
    if((d->flags & C_CONNECTED) && is_on_channel(d->player, channel) &&
       (controls(player, d->player, POW_WHO) ||
       could_doit(player, d->player, A_HLOCK))) {
      notify(player, "%s is on channel %s, idle %s.",
             unparse_object(player,d->player), channel,
             time_format(TMA, getidletime(d->player)));
      count++;
    }

  /* result count */
  notify(player, "There %s %d player%s on channel '%s'.",
         count==1?"is":"are", count, count==1?"":"s", channel);
}

/* channel-emit */
void do_cemit(dbref player, char *channel, char *msg)
{
  if(!*msg) {
    notify(player, "No message.");
    return;
  }

  default_color=unparse_color(player, player, 14);
  com_send(player, channel, msg);
  default_color=7;

  if(!Quiet(player) && !is_on_channel(player, channel))
    notify(player, "Channel '%s' just saw: '%s'.", channel, msg);
}

/* channel comlink */
void do_com(dbref player, char *arg1, char *arg2)
{
  char chan[16], message[16384], *s, *t=NULL;
  dbref thing;

  /* Use default channel if none specified */
  if(!*arg1) {
    strmaxcpy(chan, atr_get(player, A_CHANNEL), 16);
    if((s=strchr(chan, ' ')))
      *s='\0';
    arg1=chan;
  }

  /* Check if player is on this channel */
  if(Typeof(player) != TYPE_PLAYER) {
    notify(player, "Non-players can not send +com messsages.");
    return;
  } if(IS(db[player].location, TYPE_ROOM, ROOM_XZONE)) {
    notify(player, "Your communicator doesn't seem to pick up a frequency.");
    return;
  } if(!*arg1) {
    notify(player, "You aren't on any channels!");
    notify(player, "To join a general discussion channel, type: +ch +public");
    return;
  } if(!is_on_channel(player, arg1)) {
    notify(player, "You aren't on channel '%s'.", arg1);
    return;
  } if(!strcasecmp(arg2, "who")) {
    com_who(player, arg1);
    return;
  } if(!charge_money(player, COM_COST, 1))
    return;

  /* Compose message using comtitle or to-recipient method */
  default_color=unparse_color(player, player, 14);
  s=message+sprintf(message, "%s", colorname(player));

  /* Check for a to-recipient */
  if(*arg2 == '\'' && (t=strchr(arg2, ' '))) {
    *t='\0';
    if(*(arg2+1) && (thing=lookup_player(arg2+1)) > NOTHING &&
       (db[thing].flags & PLAYER_CONNECT) && (controls(player, thing, POW_WHO)
       || could_doit(player, thing, A_HLOCK))) {
      s+=sprintf(s, " [to %s]", colorname(thing));
      arg2=t?t+1:"";
      if(*arg2 == ';')
        s+=sprintf(s, ": %s", colorname(player));
    } else {
      *t=' ';
      t=NULL;
    }
  }

  /* Append player's comtitle */
  if(!t && *arg2 != ':' && *arg2 != ';' &&
     *(t=atr_parse(player, player, A_COMTITLE)))
    s+=sprintf(s, " <%s>", t);

  /* Append text to message */
  sprintf(s, "%s %s", (*arg2 == ':')?"":((*arg2 == ';')?
          possname(db[player].name):":"), (*arg2 == ':' || *arg2 == ';')?
          arg2+1:arg2);
  message[8000]='\0';

  /* Send it */
  com_send(player, arg1, message);

  /* Restore defualt color */
  default_color=7;
}

void do_channel(dbref player, char *arg1)
{
  char buf[8192], *s;
  int i;
  
  if(Typeof(player) != TYPE_PLAYER) {
    notify(player, "Objects may not use channels.");
    return;
  } if(strchr(arg1, ' ')) {
    notify(player, "Sorry, channel names may not have spaces in them.");
    return;
  }

  /* list channels? */
  if(!*arg1) {
    if(*(s=atr_get(player, A_CHANNEL)))
      notify(player, "You are currently on the following channels: %s", s);
    else {
      notify(player, "You are currently not on any channels.");
      notify(player, "For a general chatting session, try: +ch +public");
    }
    return;
  }
  
  /* delete channel */
  if(*arg1 == '-') {
    if(!*++arg1) {
      notify(player, "Usage: +ch -<channel>");
      return;
    } if(!(i=is_on_channel(player, arg1))) {
      notify(player, "You are not on that channel.");
      return;
    }

    default_color=unparse_color(player, player, 14);
    com_send(player, arg1, tprintf("* %s has left this channel.",
             colorname(player)));
    default_color=7;

    /* remove channel from list */
    s=atr_get(player, A_CHANNEL);
    strmaxcpy(buf, s, i);
    if((s=strchr(s+i, ' ')))
      strcat(buf, s+1);
    else if(i > 1)
      buf[i-2]='\0';
    atr_add(player, A_CHANNEL, buf);
    return;
  }

  if(*arg1 == '+') {  /* join channel */
    if(!*++arg1) {
      notify(player, "Usage: +ch +<channel>");
      return;
    } if(is_on_channel(player, arg1)) {
      notify(player, "You are already on that channel.");
      return;
    }
  }

  if(strlen(arg1) > 15) {
    notify(player, "Channel names may only be 15 characters long.");
    return;
  }

  /* set default channel */
  if((i=is_on_channel(player, arg1))) {
    s=atr_get(player, A_CHANNEL);
    strmaxcpy(buf, s, i);
    if((s=strchr(s+i, ' ')))
      strcat(buf, s+1);
    else if(i > 1)
      buf[i-2]='\0';
    atr_add(player, A_CHANNEL, tprintf("%s%s%s", arg1, *buf?" ":"", buf));
  } else {
    if(IS(db[player].location, TYPE_ROOM, ROOM_XZONE)) {
      notify(player, "Can't join channels from here.");
      return;
    } if(!ok_channel_change(player, arg1)) {
      notify(player, "You have no access to that channel.");
      return;
    }

    s=atr_get(player, A_CHANNEL);
    if(strlen(s)+strlen(arg1) > 7999) {
      notify(player, "Error: Too many channels joined.");
      return;
    }

    atr_add(player, A_CHANNEL, tprintf("%s%s%s", arg1, *s?" ":"", s));
    default_color=unparse_color(player, player, 14);
    com_send(player, arg1, tprintf("* %s has joined this channel.",
             colorname(player)));
    default_color=7;
  }

  notify(player, "Channel %s is now default.", arg1);
}

int ok_channel_change(dbref player, char *channel)
{
  /* check permission to join admin-restricted channels */
  if(power(player, POW_SECURITY))
    return 1;
  if(*channel == '*' || (*channel == '.' && !power(player, POW_WHO)) ||
     (*channel == '_' && !Immortal(player)))
    return 0;

  if(!power(player, POW_WHO) && (*channel == '%' || *channel == '$'))
    return 0;
  return 1;
}

/* removes all channels beginning with the symbol 'type' */
void remove_channel_type(dbref player, char type)
{
  char buf[8192], *r, *s=atr_get(player, A_CHANNEL), *t=buf;

  *buf='\0';
  while((r=parse_up(&s,' ')))
    if(*r != type) {
      if(*buf)
        *t++=' ';
      t+=sprintf(t, "%s", r);
    }

  atr_add(player, A_CHANNEL, buf);
}
